# Copyright 2022 QP Hou
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List

from rp2.plugin.country.us import US
from rp2.rp2_error import RP2RuntimeError

from dali.in_transaction import InTransaction
from dali.out_transaction import OutTransaction
from dali.plugin.input.rest.coinbase_pro import InputPlugin


class TestSwapFill:
    def test_buy_side(self, mocker: Any) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            api_key="a",
            api_secret="b",
            api_passphrase="c",
            native_fiat="USD",
            thread_count=1,
        )

        btc_account_id = "bbbbbbbb-dddd-4444-8888-000000000000"
        eth_account_id = "eeeeeeee-4444-5555-aaaa-cccccccccccc"
        mocker.patch.object(plugin, "_InputPlugin__get_accounts").return_value = [
            {
                "id": btc_account_id,
                "currency": "BTC",
            },
            {
                "id": eth_account_id,
                "currency": "ETH",
            },
        ]

        def mock_get_transaction(account_id: str) -> List[Dict[str, Any]]:
            if account_id == eth_account_id:
                return [
                    {
                        "id": "1111111111",
                        "amount": "10.0000000000000000",
                        "balance": "11.0000000000000000",
                        "created_at": "2020-12-11T00:20:14.693676Z",
                        "type": "match",
                        "details": {
                            "order_id": "33333333-bbbb-4444-cccc-aaaaaaaaaaaa",
                            "product_id": "ETH-BTC",
                            "trade_id": "134567890",
                        },
                    },
                ]

            if account_id == btc_account_id:
                return [
                    {
                        "id": "1111111111",
                        "amount": "-0.0005000000000000",
                        "balance": "0.9995000000000000",
                        "created_at": "2020-12-11T00:20:14.693676Z",
                        "type": "fee",
                        "details": {
                            "order_id": "33333333-bbbb-4444-cccc-aaaaaaaaaaaa",
                            "product_id": "ETH-BTC",
                            "trade_id": "134567890",
                        },
                    },
                    {
                        "id": "1111111108",
                        "amount": "-0.5000000000000000",
                        "balance": "1.0000000000000000",
                        "created_at": "2020-12-11T00:20:14.693676Z",
                        "type": "match",
                        "details": {
                            "order_id": "33333333-bbbb-4444-cccc-aaaaaaaaaaaa",
                            "product_id": "ETH-BTC",
                            "trade_id": "134567890",
                        },
                    },
                ]

            raise RP2RuntimeError("Invalid account id: " + account_id)

        mocker.patch.object(plugin, "_InputPlugin__get_transactions").side_effect = mock_get_transaction

        mocker.patch.object(plugin, "_InputPlugin__get_fills").return_value = [
            {
                "created_at": "2020-12-11T00:20:14.693676Z",
                "trade_id": "134567890",
                "product_id": "ETH-BTC",
                "order_id": "33333333-bbbb-4444-cccc-aaaaaaaaaaaa",
                "liquidity": "M",
                "price": "0.05000000",
                "size": "10.00000000",
                "fee": "0.0005000000000000",
                "side": "buy",
                "settled": True,
                "usd_volume": "25000.000000000000000000000000",
            }
        ]

        result = plugin.load(US())
        assert len(result) == 2

        in_transaction: InTransaction = result[0]  # type: ignore
        out_transaction: OutTransaction = result[1]  # type: ignore

        assert out_transaction.asset == "BTC"
        assert out_transaction.timestamp == "2020-12-11 00:20:14.693676+0000"
        assert out_transaction.transaction_type == "Sell"
        assert out_transaction.spot_price == "50000.00000000"
        assert out_transaction.crypto_out_no_fee == "0.5000000000000000"
        assert out_transaction.crypto_fee == "0.0005000000000000"
        assert out_transaction.crypto_out_with_fee is None
        assert out_transaction.fiat_out_no_fee is None
        assert out_transaction.fiat_fee is None

        assert in_transaction.asset == "ETH"
        assert in_transaction.timestamp == "2020-12-11 00:20:14.693676+0000"
        assert in_transaction.transaction_type == "Buy"
        assert in_transaction.spot_price == "2500.0000000000000000"
        assert in_transaction.crypto_in == "10.00000000"
        assert in_transaction.crypto_fee == "0"
        assert in_transaction.fiat_in_no_fee is None
        assert in_transaction.fiat_in_with_fee is None
        assert in_transaction.fiat_fee is None



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Bitbank REST plugin links:
# REST API: https://docs.bitbank.cc/
# Authentication: https://github.com/bitbankinc/bitbank-api-docs/blob/master/rest-api.md#authorization
# Endpoint: https://bitbank.cc/

# CCXT documentation:
# https://docs.ccxt.com/en/latest/index.html

import json
from datetime import datetime
from typing import Any, List, Optional

from ccxt import bitbank

from rp2.rp2_decimal import ZERO, RP2Decimal

from dali.abstract_ccxt_input_plugin import (
    AbstractCcxtInputPlugin,
    ProcessOperationResult,
)
from dali.ccxt_pagination import (
    AbstractPaginationDetailSet,
    DateBasedPaginationDetailSet,
)
from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction

# Time period constants
_MS_IN_SECOND: int = 1000

# Record limits
_TRADE_RECORD_LIMIT: int = 1000

# Transaction keywords
_COST: str = "cost"
_CURRENCY: str = "currency"
_FEE: str = "fee"
_ID: str = "id"
_TIMESTAMP: str = "timestamp"


class InputPlugin(AbstractCcxtInputPlugin):

    __EXCHANGE_NAME: str = "Bitbank.cc"
    __PLUGIN_NAME: str = "Bitbank.cc_REST"
    __DEFAULT_THREAD_COUNT: int = 1

    def __init__(
        self,
        account_holder: str,
        api_key: str,
        api_secret: str,
        native_fiat: str,
        thread_count: Optional[int] = __DEFAULT_THREAD_COUNT,
    ) -> None:

        self.__api_key = api_key
        self.__api_secret = api_secret
        # We will have a default start time of March 1st, 2017 since Bitbank Exchange officially launched on March 1st Japan Time.
        super().__init__(account_holder, datetime(2017, 3, 1, 0, 0, 0, 0), native_fiat, thread_count)

    def exchange_name(self) -> str:
        return self.__EXCHANGE_NAME

    def plugin_name(self) -> str:
        return self.__PLUGIN_NAME

    def _initialize_client(self) -> bitbank:
        return bitbank(
            {
                "apiKey": self.__api_key,
                "enableRateLimit": True,
                "secret": self.__api_secret,
            }
        )

    def _get_process_deposits_pagination_detail_set(self) -> Optional[AbstractPaginationDetailSet]:
        return None

    def _get_process_withdrawals_pagination_detail_set(self) -> Optional[AbstractPaginationDetailSet]:
        return None

    def _get_process_trades_pagination_detail_set(self) -> Optional[AbstractPaginationDetailSet]:
        return DateBasedPaginationDetailSet(
            limit=_TRADE_RECORD_LIMIT,
            exchange_start_time=self._start_time_ms,
            markets=self._get_markets(),
        )

    def _process_gains(
        self,
        in_transactions: List[InTransaction],
        out_transactions: List[OutTransaction],
    ) -> None:
        pass

    def _process_implicit_api(
        self,
        in_transactions: List[InTransaction],
        out_transactions: List[OutTransaction],
        intra_transactions: List[IntraTransaction],
    ) -> None:
        pass

    # Override the handling of trades since Bitbank has negative maker fees
    def _process_buy_and_sell(self, transaction: Any, notes: Optional[str] = None) -> ProcessOperationResult:
        fee_income: Optional[InTransaction] = None
        if RP2Decimal(str(transaction[_FEE][_COST])) < ZERO:
            fee_income = InTransaction(
                plugin=self.plugin_name(),
                unique_id=f"{transaction[_ID]}",
                raw_data=json.dumps(transaction),
                timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_TIMESTAMP]),
                asset=transaction[_FEE][_CURRENCY],
                exchange=self.exchange_name(),
                holder=self.account_holder,
                transaction_type=Keyword.INCOME.value,
                spot_price=Keyword.UNKNOWN.value,
                crypto_in=str(-transaction[_FEE][_COST]),
                crypto_fee=None,
                fiat_in_no_fee=None,
                fiat_in_with_fee=None,
                fiat_fee=None,
                notes=(f"{notes + '; ' if notes else ''} Fee income for transaction #{transaction[_ID]}"),
            )

            # Zero out the fee so it is not added as a negative fee
            transaction[_FEE][_COST] = 0.0

        results: ProcessOperationResult = super()._process_buy_and_sell(transaction, notes)
        if fee_income:
            results.in_transactions.append(fee_income)

        return results



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal

from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.plugin.input.csv.coincheck_supplemental import InputPlugin


class TestCoincheck:
    def test_trades(self) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            buys_csv_file="input/test_coincheck_buys.csv",
            native_fiat="USD",
        )

        result = plugin.load(US())

        # 1 0.005 BTC purchase
        assert len(result) == 1

        btc_buy_in: InTransaction = result[0]  # type: ignore

        assert btc_buy_in.asset == "BTC"
        assert btc_buy_in.timestamp == "2022-04-20 16:20:00+0000"
        assert btc_buy_in.transaction_type == Keyword.BUY.value.capitalize()
        assert RP2Decimal(str(btc_buy_in.spot_price)) == RP2Decimal("2000000")
        assert RP2Decimal(btc_buy_in.crypto_in) == RP2Decimal("0.005")
        assert btc_buy_in.crypto_fee is None
        assert RP2Decimal(str(btc_buy_in.fiat_in_no_fee)) == RP2Decimal("10000")
        assert RP2Decimal(str(btc_buy_in.fiat_in_with_fee)) == RP2Decimal("10000")
        assert btc_buy_in.fiat_fee is None



# Copyright 2023 Neal Chambers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from dali.plugin.pair_converter.ccxt import (
    PairConverterPlugin as CcxtPairConverterPlugin,
)


class PairConverterPlugin(CcxtPairConverterPlugin):
    def __init__(
        self,
        historical_price_type: str,
        fiat_priority: Optional[str] = None,
        google_api_key: Optional[str] = None,
    ) -> None:

        super().__init__(
            historical_price_type=historical_price_type,
            default_exchange="Kraken",
            fiat_priority=fiat_priority,
            google_api_key=google_api_key,
            exchange_locked=True,
        )



# Copyright 2022 mbianco
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# CSV Format: transaction id, type, currency, amount, usd equivalent, details, outstanding loan, datetime

import logging
import re
from csv import reader
from decimal import Decimal
from typing import List, Optional

from rp2.abstract_country import AbstractCountry
from rp2.logger import create_logger

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction

# transaction types
# locking & unlocking deposits are transfers within Nexo. They cannot lock funds from other wallets/accounts.
_INTEREST = "Interest"
_LOCKING_TERM_DEPOSIT = "LockingTermDeposit"
_UNLOCKING_TERM_DEPOSIT = "UnlockingTermDeposit"
_FIXED_TERM_INTEREST = "FixedTermInterest"
_DEPOSIT = "Deposit"


class InputPlugin(AbstractInputPlugin):
    __NEXO: str = "Nexo"

    __TRANSACTION_TYPE_INDEX = 1
    __CURRENCY_INDEX: int = 2
    __AMOUNT_INDEX: int = 3
    __SPOT_PRICE_INDEX: int = 4
    __DETAILS_INDEX = 5

    __TIMESTAMP_INDEX: int = 7

    __DELIMITER = ","

    def __init__(
        self,
        account_holder: str,
        transaction_csv_file: str,
        native_fiat: Optional[str] = None,
    ) -> None:

        super().__init__(account_holder=account_holder, native_fiat=native_fiat)
        self.__transaction_csv_file: str = transaction_csv_file
        self.__logger: logging.Logger = create_logger(f"{self.__NEXO}/{self.account_holder}")

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        with open(self.__transaction_csv_file, encoding="utf-8") as transaction_csv_file:
            # read CSV with header and skip first row
            lines = reader(transaction_csv_file)

            # Skip header line
            header = next(lines)
            self.__logger.debug("Header: %s", header)

            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                self.__logger.debug("Transaction: %s", raw_data)

                transaction_type: str = line[self.__TRANSACTION_TYPE_INDEX].strip()
                currency: str = line[self.__CURRENCY_INDEX].strip()
                amount = line[self.__AMOUNT_INDEX].strip()
                # there is no timezone information in the CSV, so we assume UTC
                timestamp_with_timezone = f"{line[self.__TIMESTAMP_INDEX].strip()} -00:00"

                common_params = {
                    # although there is a transaction id in the CSV, it is not a transaction hash shared across exchanges, so it is uselsss
                    # https://github.com/eprbell/dali-rp2/pull/60#issuecomment-1201481064
                    "unique_id": Keyword.UNKNOWN.value,
                    "plugin": self.__NEXO,
                    "raw_data": raw_data,
                    "timestamp": timestamp_with_timezone,
                    "asset": currency,
                }

                # nexo gives us the realized usd, but it's often 0 if a subcent value
                # if it is non-zero, we use it to calculate the spot price, otherwise we use unknown as the value
                # the spot price contains $ char, so we remove it. Sometimes the price is specified with scientific notation,
                # so we are not using a more general regex like `[^\d.]`
                realized_usd = Decimal(re.sub(r"[$]", "", line[self.__SPOT_PRICE_INDEX]))
                spot_price = str(realized_usd / Decimal(amount)) if not realized_usd.is_zero() else Keyword.UNKNOWN.value

                if transaction_type in [_INTEREST, _FIXED_TERM_INTEREST]:
                    result.append(
                        InTransaction(
                            **(
                                common_params  # type: ignore
                                | {
                                    "exchange": self.__NEXO,
                                    "holder": self.account_holder,
                                    "transaction_type": Keyword.INTEREST.value,
                                    "spot_price": spot_price,
                                    "crypto_in": amount,
                                    "fiat_fee": "0",
                                }
                            )
                        )
                    )
                elif transaction_type in [_LOCKING_TERM_DEPOSIT, _UNLOCKING_TERM_DEPOSIT]:
                    # These are unique to Nexo: they "lock" your crypto in a "fixed term" deposit which earns higher interest.
                    # i.e. these transactions just indicate that you cannot withdraw these funds while these are locked. So they effect your available balance.
                    # I don't think we need to record locking/unlocking deposits for term interest
                    self.__logger.debug("Skipping lock or unlock deposit: %s", line)
                elif transaction_type == _DEPOSIT:
                    unique_id = common_params["unique_id"]

                    # nexo includes a network transaction hash in the case of deposits: it's embedded in the details field
                    # let's extract it using a regex
                    if transaction_hash_match := re.search(r"([A-Fa-f0-9]{64})$", line[self.__DETAILS_INDEX]):
                        unique_id = transaction_hash_match[0]

                    result.append(
                        IntraTransaction(
                            **(
                                common_params  # type: ignore
                                | {
                                    "unique_id": unique_id,
                                    "crypto_received": amount,
                                    # most likely, funds are coming from the user/tax payer, but we can't say for sure so we use unknown
                                    # and let the DaLI transaction resolver fill in the missing details.
                                    "from_exchange": Keyword.UNKNOWN.value,
                                    "from_holder": Keyword.UNKNOWN.value,
                                    "to_exchange": self.__NEXO,
                                    "to_holder": self.account_holder,
                                    # we do know the spot price, but nexo seems to do some aggressive rounding
                                    "spot_price": spot_price,
                                    "crypto_sent": Keyword.UNKNOWN.value,
                                }
                            )
                        )
                    )
                else:
                    # TODO in my data, I had no withdrawals, they will need to be implemented in the future # pylint: disable=fixme
                    self.__logger.error("Unsupported transaction type (skipping): %s. Please open an issue at %s", raw_data, self.ISSUES_URL)

        return result



# Copyright 2022 Neal Chambers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Union

from ccxt import binance, kraken
from rp2.rp2_decimal import ZERO, RP2Decimal

from dali.abstract_pair_converter_plugin import AssetPairAndTimestamp
from dali.cache import CACHE_DIR, load_from_cache
from dali.configuration import Keyword
from dali.historical_bar import HistoricalBar
from dali.plugin.pair_converter.ccxt import PairConverterPlugin
from dali.plugin.pair_converter.csv.kraken import Kraken as KrakenCsvPricing

# Default exchange
TEST_EXCHANGE: str = "Kraken"
ALT_EXCHANGE: str = "Binance.com"
LOCKED_EXCHANGE: str = "Kraken"
FIAT_EXHANGE: str = "fiat"
TEST_GRAPH: Dict[str, List[str]] = {
    "BETH": ["ETH"],
    "BTC": ["USDT", "GBP"],
    "ETH": ["USDT"],
    "USDT": ["USD"],
    "USD": ["JPY"],
}
TEST_MARKETS: Dict[str, List[str]] = {
    "BTCUSDT": [ALT_EXCHANGE],
    "BTCGBP": [ALT_EXCHANGE],
    "BETHETH": [ALT_EXCHANGE],
    "ETHUSDT": [ALT_EXCHANGE],
    "USDTUSD": [TEST_EXCHANGE],
    "USDJPY": [FIAT_EXHANGE],
}
LOCKED_MARKETS: Dict[str, List[str]] = {
    "BTCUSDT": [TEST_EXCHANGE],
    "USDTUSD": [TEST_EXCHANGE],
}

# BTCUSDT conversion
BAR_DURATION: str = "1m"
BAR_TIMESTAMP: datetime = datetime.fromtimestamp(1504541580, timezone.utc)
BAR_LOW: RP2Decimal = RP2Decimal("4230.0")
BAR_HIGH: RP2Decimal = RP2Decimal("4240.6")
BAR_OPEN: RP2Decimal = RP2Decimal("4235.4")
BAR_CLOSE: RP2Decimal = RP2Decimal("4230.7")
BAR_VOLUME: RP2Decimal = RP2Decimal("37.72941911")

# USDTUSD conversion
USDTUSD_DURATION: str = "1m"
USDTUSD_TIMESTAMP: datetime = datetime.fromtimestamp(1504541580, timezone.utc)
USDTUSD_LOW: RP2Decimal = RP2Decimal("0.9987")
USDTUSD_HIGH: RP2Decimal = RP2Decimal("0.9988")
USDTUSD_OPEN: RP2Decimal = RP2Decimal("0.9987")
USDTUSD_CLOSE: RP2Decimal = RP2Decimal("0.9988")
USDTUSD_VOLUME: RP2Decimal = RP2Decimal("113.786789")

# Missing Fiat Test
JPY_USD_RATE: RP2Decimal = RP2Decimal("115")

# No Fiat pair conversion
BETHETH_TIMESTAMP: datetime = datetime.fromtimestamp(1504541590, timezone.utc)
BETHETH_LOW: RP2Decimal = RP2Decimal("0.9722")
BETHETH_HIGH: RP2Decimal = RP2Decimal("0.9739")
BETHETH_OPEN: RP2Decimal = RP2Decimal("0.9736")
BETHETH_CLOSE: RP2Decimal = RP2Decimal("0.9735")
BETHETH_VOLUME: RP2Decimal = RP2Decimal("309")

ETHUSDT_LOW: RP2Decimal = RP2Decimal("1753.75")
ETHUSDT_HIGH: RP2Decimal = RP2Decimal("1764.70")
ETHUSDT_OPEN: RP2Decimal = RP2Decimal("1755.81")
ETHUSDT_CLOSE: RP2Decimal = RP2Decimal("1763.03")
ETHUSDT_VOLUME: RP2Decimal = RP2Decimal("434")

# Non-USD Fiat pair conversion
BTCGBP_TIMESTAMP: datetime = datetime.fromtimestamp(1504541600, timezone.utc)
BTCGBP_LOW: RP2Decimal = RP2Decimal("3379.06")
BTCGBP_HIGH: RP2Decimal = RP2Decimal("3387.53")
BTCGBP_OPEN: RP2Decimal = RP2Decimal("3383.29")
BTCGBP_CLOSE: RP2Decimal = RP2Decimal("3379.66")
BTCGBP_VOLUME: RP2Decimal = RP2Decimal("37.72941911")

# Fiat to Fiat Test
EUR_USD_RATE: RP2Decimal = RP2Decimal("1.0847")
EUR_USD_TIMESTAMP: datetime = datetime.fromtimestamp(1585958400, timezone.utc)

# Kraken CSV read Test
KRAKEN_TIMESTAMP: datetime = datetime.fromtimestamp(1601855760, timezone.utc)
KRAKEN_LOW: RP2Decimal = RP2Decimal("1.5557")
KRAKEN_HIGH: RP2Decimal = RP2Decimal("1.5556")
KRAKEN_OPEN: RP2Decimal = RP2Decimal("1.5555")
KRAKEN_CLOSE: RP2Decimal = RP2Decimal("1.5558")
KRAKEN_VOLUME: RP2Decimal = RP2Decimal("15.15")

_MS_IN_SECOND: int = 1000


class TestCcxtPlugin:
    def __btcusdt_mock(self, plugin: PairConverterPlugin, mocker: Any) -> None:
        exchange = kraken(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )
        alt_exchange = binance(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )

        kraken_csv = KrakenCsvPricing(google_api_key="whatever")

        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_markets", {TEST_EXCHANGE: TEST_MARKETS})
        mocker.patch.object(kraken_csv, "find_historical_bar", [])
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_csv_reader", {"kraken": kraken_csv})
        mocker.patch.object(alt_exchange, "fetchOHLCV").return_value = [
            [
                BAR_TIMESTAMP,  # UTC timestamp in milliseconds, integer
                BAR_OPEN,  # (O)pen price, float
                BAR_HIGH,  # (H)ighest price, float
                BAR_LOW,  # (L)owest price, float
                BAR_CLOSE,  # (C)losing price, float
                BAR_VOLUME,  # (V)olume (in terms of the base currency), float
            ],
        ]

        mocker.patch.object(exchange, "fetchOHLCV").return_value = [
            [
                USDTUSD_TIMESTAMP,  # UTC timestamp in milliseconds, integer
                USDTUSD_OPEN,  # (O)pen price, float
                USDTUSD_HIGH,  # (H)ighest price, float
                USDTUSD_LOW,  # (L)owest price, float
                USDTUSD_CLOSE,  # (C)losing price, float
                USDTUSD_VOLUME,  # (V)olume (in terms of the base currency), float
            ],
        ]
        mocker.patch.object(plugin, "_PairConverterPlugin__exchanges", {TEST_EXCHANGE: exchange, ALT_EXCHANGE: alt_exchange})
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_graphs", {TEST_EXCHANGE: TEST_GRAPH})

    def test_unknown_exchange(self, mocker: Any) -> None:
        plugin: PairConverterPlugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value)
        self.__btcusdt_mock(plugin, mocker)

        data = plugin.get_historic_bar_from_native_source(BAR_TIMESTAMP, "BTC", "USD", "Bogus Exchange")
        assert data

    def test_historical_prices(self, mocker: Any) -> None:
        plugin: PairConverterPlugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value)
        cache_path = os.path.join(CACHE_DIR, plugin.cache_key())
        if os.path.exists(cache_path):
            os.remove(cache_path)

        # Reinstantiate plugin now that cache is gone
        plugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value)
        self.__btcusdt_mock(plugin, mocker)

        data = plugin.get_historic_bar_from_native_source(BAR_TIMESTAMP, "BTC", "USD", TEST_EXCHANGE)

        assert data
        assert data.timestamp == BAR_TIMESTAMP
        assert data.low == BAR_LOW * USDTUSD_LOW
        assert data.high == BAR_HIGH * USDTUSD_HIGH
        assert data.open == BAR_OPEN * USDTUSD_OPEN
        assert data.close == BAR_CLOSE * USDTUSD_CLOSE
        assert data.volume == BAR_VOLUME + USDTUSD_VOLUME

        # Read price again, but populate plugin cache this time
        value = plugin.get_conversion_rate(BAR_TIMESTAMP, "BTC", "USD", TEST_EXCHANGE)
        assert value
        assert value == BAR_HIGH * USDTUSD_HIGH

        # Save plugin cache
        plugin.save_historical_price_cache()

        # Load plugin cache and verify
        cache = load_from_cache(plugin.cache_key())
        key = AssetPairAndTimestamp(BAR_TIMESTAMP, "BTC", "USD", TEST_EXCHANGE)

        # 3 cached prices - BTC/USDT, USDT/USD, BTC/USD
        assert len(cache) == 3, str(cache)
        assert key in cache
        data = cache[key]

        assert data
        assert data.timestamp == BAR_TIMESTAMP
        assert data.timestamp == BAR_TIMESTAMP
        assert data.low == BAR_LOW * USDTUSD_LOW
        assert data.high == BAR_HIGH * USDTUSD_HIGH
        assert data.open == BAR_OPEN * USDTUSD_OPEN
        assert data.close == BAR_CLOSE * USDTUSD_CLOSE
        assert data.volume == BAR_VOLUME + USDTUSD_VOLUME

    def test_missing_historical_prices(self, mocker: Any) -> None:
        plugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value)
        timestamp = datetime(2020, 6, 1, 0, 0)

        mocker.patch.object(plugin, "get_historic_bar_from_native_source").return_value = None

        data = plugin.get_historic_bar_from_native_source(timestamp, "BOGUSCOIN", "JPY", TEST_EXCHANGE)
        assert data is None

    def test_missing_fiat_pair(self, mocker: Any) -> None:
        plugin: PairConverterPlugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value)
        self.__btcusdt_mock(plugin, mocker)

        mocker.patch.object(plugin, "_get_fiat_exchange_rate").return_value = HistoricalBar(
            duration=timedelta(seconds=86400),
            timestamp=BAR_TIMESTAMP,
            open=RP2Decimal(str(JPY_USD_RATE)),
            high=RP2Decimal(str(JPY_USD_RATE)),
            low=RP2Decimal(str(JPY_USD_RATE)),
            close=RP2Decimal(str(JPY_USD_RATE)),
            volume=ZERO,
        )

        data = plugin.get_historic_bar_from_native_source(BAR_TIMESTAMP, "BTC", "JPY", TEST_EXCHANGE)

        assert data
        assert data.timestamp == BAR_TIMESTAMP
        assert data.low == BAR_LOW * USDTUSD_LOW * JPY_USD_RATE
        assert data.high == BAR_HIGH * USDTUSD_HIGH * JPY_USD_RATE
        assert data.open == BAR_OPEN * USDTUSD_OPEN * JPY_USD_RATE
        assert data.close == BAR_CLOSE * USDTUSD_CLOSE * JPY_USD_RATE
        assert data.volume == BAR_VOLUME + USDTUSD_VOLUME

    # Some crypto assets have no fiat or stable coin pair; they are only paired with BTC or ETH (e.g. EZ or BETH)
    # To get an accurate fiat price, we must get the price in the base asset (e.g. BETH -> ETH) then convert that to fiat (e.g. ETH -> USD)
    def test_no_fiat_pair(self, mocker: Any) -> None:
        plugin: PairConverterPlugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value)

        exchange = kraken(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )
        alt_exchange = binance(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_markets", {TEST_EXCHANGE: TEST_MARKETS})

        def no_fiat_fetch_ohlcv(symbol: str, timeframe: str, timestamp: int, candles: int) -> List[List[Union[float, int]]]:
            # pylint: disable=unused-argument
            result: List[List[Union[float, int]]] = []
            if symbol == "BETH/ETH":
                result = [
                    [
                        BETHETH_TIMESTAMP.timestamp() * _MS_IN_SECOND,  # UTC timestamp in milliseconds, integer
                        float(BETHETH_OPEN),  # (O)pen price, float
                        float(BETHETH_HIGH),  # (H)ighest price, float
                        float(BETHETH_LOW),  # (L)owest price, float
                        float(BETHETH_CLOSE),  # (C)losing price, float
                        float(BETHETH_VOLUME),  # (V)olume (in terms of the base currency), float
                    ],
                ]
            elif symbol == "ETH/USDT":
                result = [
                    [
                        BETHETH_TIMESTAMP.timestamp() * _MS_IN_SECOND,  # UTC timestamp in milliseconds, integer
                        float(ETHUSDT_OPEN),  # (O)pen price, float
                        float(ETHUSDT_HIGH),  # (H)ighest price, float
                        float(ETHUSDT_LOW),  # (L)owest price, float
                        float(ETHUSDT_CLOSE),  # (C)losing price, float
                        float(ETHUSDT_VOLUME),  # (V)olume (in terms of the base currency), float
                    ],
                ]

            return result

        mocker.patch.object(alt_exchange, "fetchOHLCV").side_effect = no_fiat_fetch_ohlcv
        mocker.patch.object(exchange, "fetchOHLCV").return_value = [
            [
                USDTUSD_TIMESTAMP,  # UTC timestamp in milliseconds, integer
                USDTUSD_OPEN,  # (O)pen price, float
                USDTUSD_HIGH,  # (H)ighest price, float
                USDTUSD_LOW,  # (L)owest price, float
                USDTUSD_CLOSE,  # (C)losing price, float
                USDTUSD_VOLUME,  # (V)olume (in terms of the base currency), float
            ],
        ]
        mocker.patch.object(plugin, "_PairConverterPlugin__exchanges", {TEST_EXCHANGE: exchange, ALT_EXCHANGE: alt_exchange})
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_graphs", {TEST_EXCHANGE: TEST_GRAPH})

        data = plugin.get_historic_bar_from_native_source(BETHETH_TIMESTAMP, "BETH", "USD", TEST_EXCHANGE)

        assert data
        assert data.timestamp == BETHETH_TIMESTAMP
        assert data.low == BETHETH_LOW * ETHUSDT_LOW * USDTUSD_LOW
        assert data.high == BETHETH_HIGH * ETHUSDT_HIGH * USDTUSD_HIGH
        assert data.open == BETHETH_OPEN * ETHUSDT_OPEN * USDTUSD_OPEN
        assert data.close == BETHETH_CLOSE * ETHUSDT_CLOSE * USDTUSD_CLOSE
        assert data.volume == BETHETH_VOLUME + ETHUSDT_VOLUME + USDTUSD_VOLUME

    # Test to make sure the default stable coin is not used with a fiat market that does exist on the exchange
    def test_nonusd_fiat_pair(self, mocker: Any) -> None:
        plugin: PairConverterPlugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value, default_exchange="Binance.com")
        alt_exchange = binance(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )
        exchange = kraken(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_markets", {TEST_EXCHANGE: TEST_MARKETS})
        mocker.patch.object(exchange, "fetchOHLCV").return_value = [
            [
                BAR_TIMESTAMP,  # UTC timestamp in milliseconds, integer
                BAR_OPEN,  # (O)pen price, float
                BAR_HIGH,  # (H)ighest price, float
                BAR_LOW,  # (L)owest price, float
                BAR_CLOSE,  # (C)losing price, float
                BAR_VOLUME,  # (V)olume (in terms of the base currency), float
            ],
        ]
        mocker.patch.object(alt_exchange, "fetchOHLCV").return_value = [
            [
                BTCGBP_TIMESTAMP,  # UTC timestamp in milliseconds, integer
                BTCGBP_OPEN,  # (O)pen price, float
                BTCGBP_HIGH,  # (H)ighest price, float
                BTCGBP_LOW,  # (L)owest price, float
                BTCGBP_CLOSE,  # (C)losing price, float
                BTCGBP_VOLUME,  # (V)olume (in terms of the base currency), float
            ],
        ]
        mocker.patch.object(plugin, "_PairConverterPlugin__exchanges", {TEST_EXCHANGE: exchange, ALT_EXCHANGE: alt_exchange})
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_graphs", {TEST_EXCHANGE: TEST_GRAPH})

        data = plugin.get_historic_bar_from_native_source(BTCGBP_TIMESTAMP, "BTC", "GBP", TEST_EXCHANGE)

        assert data
        assert data.timestamp == BTCGBP_TIMESTAMP
        assert data.low == BTCGBP_LOW
        assert data.high == BTCGBP_HIGH
        assert data.open == BTCGBP_OPEN
        assert data.close == BTCGBP_CLOSE
        assert data.volume == BTCGBP_VOLUME

    # Plugin should hand off the handling of a fiat to fiat pair to the fiat converter
    def test_fiat_pair(self, mocker: Any) -> None:
        plugin: PairConverterPlugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value)
        exchange = binance(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )

        # Need to be mocked to prevent logger spam
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_markets", {TEST_EXCHANGE: ["WHATEVER"]})
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_graphs", {TEST_EXCHANGE: TEST_GRAPH})
        mocker.patch.object(plugin, "_get_fiat_exchange_rate").return_value = HistoricalBar(
            duration=timedelta(seconds=86400),
            timestamp=EUR_USD_TIMESTAMP,
            open=RP2Decimal(str(EUR_USD_RATE)),
            high=RP2Decimal(str(EUR_USD_RATE)),
            low=RP2Decimal(str(EUR_USD_RATE)),
            close=RP2Decimal(str(EUR_USD_RATE)),
            volume=ZERO,
        )
        mocker.patch.object(plugin, "_PairConverterPlugin__exchanges", {TEST_EXCHANGE: exchange})

        data = plugin.get_historic_bar_from_native_source(EUR_USD_TIMESTAMP, "EUR", "USD", TEST_EXCHANGE)

        assert data
        assert data.timestamp == EUR_USD_TIMESTAMP
        assert data.low == EUR_USD_RATE
        assert data.high == EUR_USD_RATE
        assert data.open == EUR_USD_RATE
        assert data.close == EUR_USD_RATE
        assert data.volume == ZERO

    def test_kraken_csv(self, mocker: Any) -> None:
        plugin: PairConverterPlugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value, google_api_key="whatever")

        cache_path = os.path.join(CACHE_DIR, "Test-" + plugin.cache_key())
        if os.path.exists(cache_path):
            os.remove(cache_path)

        kraken_csv = KrakenCsvPricing(google_api_key="whatever")
        mocker.patch.object(kraken_csv, "cache_key").return_value = "Test-" + kraken_csv.cache_key()
        mocker.patch.object(kraken_csv, "_Kraken__CACHE_DIRECTORY", "output/kraken_test")
        if not os.path.exists("output/kraken_test"):
            os.makedirs("output/kraken_test")
        with open("input/USD_OHLCVT_test.zip", "rb") as file:
            mocker.patch.object(kraken_csv, "_google_file_to_bytes").return_value = file.read()

        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_csv_reader", {"Kraken": kraken_csv})
        exchange = kraken(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )
        alt_exchange = binance(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )
        modified_markets = TEST_MARKETS
        modified_markets["BTCUSDT"] = [ALT_EXCHANGE]
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_markets", {TEST_EXCHANGE: modified_markets})
        mocker.patch.object(alt_exchange, "fetchOHLCV").return_value = [
            [
                KRAKEN_TIMESTAMP,  # Match the timestamp to assure correct price look up
                BAR_OPEN,  # (O)pen price, float
                BAR_HIGH,  # (H)ighest price, float
                BAR_LOW,  # (L)owest price, float
                BAR_CLOSE,  # (C)losing price, float
                BAR_VOLUME,  # (V)olume (in terms of the base currency), float
            ],
        ]

        mocker.patch.object(exchange, "fetchOHLCV").return_value = [
            [
                KRAKEN_TIMESTAMP,  # UTC timestamp in milliseconds, integer
                USDTUSD_OPEN,  # (O)pen price, float
                USDTUSD_HIGH,  # (H)ighest price, float
                USDTUSD_LOW,  # (L)owest price, float
                USDTUSD_CLOSE,  # (C)losing price, float
                USDTUSD_VOLUME,  # (V)olume (in terms of the base currency), float
            ],
        ]
        mocker.patch.object(plugin, "_PairConverterPlugin__exchanges", {TEST_EXCHANGE: exchange, ALT_EXCHANGE: alt_exchange})
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_graphs", {TEST_EXCHANGE: TEST_GRAPH})

        data = plugin.get_historic_bar_from_native_source(KRAKEN_TIMESTAMP, "BTC", "USD", TEST_EXCHANGE)

        assert data
        assert data.timestamp == KRAKEN_TIMESTAMP
        assert data.low == BAR_LOW * KRAKEN_LOW
        assert data.high == BAR_HIGH * KRAKEN_HIGH
        assert data.open == BAR_OPEN * KRAKEN_OPEN
        assert data.close == BAR_CLOSE * KRAKEN_CLOSE
        assert data.volume == BAR_VOLUME + KRAKEN_VOLUME

    def test_locked_exchange(self, mocker: Any) -> None:
        plugin: PairConverterPlugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value, default_exchange=LOCKED_EXCHANGE, exchange_locked=True)
        # Name is changed to exchange_instance to avoid conflicts with the side effect function `add_exchange_side_effect`
        exchange_instance = kraken(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )
        alt_exchange = binance(
            {
                "apiKey": "key",
                "secret": "secret",
            }
        )
        kraken_csv = KrakenCsvPricing(google_api_key="whatever")

        mocker.patch.object(kraken_csv, "find_historical_bar").return_value = None
        mocker.patch.object(plugin, "_PairConverterPlugin__exchange_csv_reader", {LOCKED_EXCHANGE: kraken_csv})
        mocker.patch.object(exchange_instance, "fetchOHLCV").return_value = [
            [
                BAR_TIMESTAMP,  # UTC timestamp in milliseconds, integer
                BAR_OPEN,  # (O)pen price, float
                BAR_HIGH,  # (H)ighest price, float
                BAR_LOW,  # (L)owest price, float
                BAR_CLOSE,  # (C)losing price, float
                BAR_VOLUME,  # (V)olume (in terms of the base currency), float
            ],
        ]

        mocker.patch.object(alt_exchange, "fetchOHLCV").return_value = [
            [
                USDTUSD_TIMESTAMP,  # UTC timestamp in milliseconds, integer
                USDTUSD_OPEN,  # (O)pen price, float
                USDTUSD_HIGH,  # (H)ighest price, float
                USDTUSD_LOW,  # (L)owest price, float
                USDTUSD_CLOSE,  # (C)losing price, float
                USDTUSD_VOLUME,  # (V)olume (in terms of the base currency), float
            ],
        ]

        def add_exchange_side_effect(exchange: str) -> None:  # pylint: disable=unused-argument
            mocker.patch.object(plugin, "_PairConverterPlugin__exchanges", {LOCKED_EXCHANGE: exchange_instance})
            mocker.patch.object(plugin, "_PairConverterPlugin__exchange_graphs", {LOCKED_EXCHANGE: TEST_GRAPH})
            mocker.patch.object(plugin, "_PairConverterPlugin__exchange_markets", {LOCKED_EXCHANGE: LOCKED_MARKETS})

        mocker.patch.object(plugin, "_add_exchange_to_memcache", autospec=True).side_effect = add_exchange_side_effect

        data = plugin.get_historic_bar_from_native_source(BAR_TIMESTAMP, "BTC", "USD", "not-kraken")

        assert data
        assert mocker.patch.object(plugin, "_add_exchange_to_memcache").called_once_with(LOCKED_EXCHANGE)



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# CSV Format: currency, amount, type, timestamp
# Note: BlockFi doesn't provide hash information, so BlockFi transactions cannot be resolved

import logging
from csv import reader
from typing import Dict, List, Optional

from rp2.abstract_country import AbstractCountry
from rp2.logger import create_logger
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2RuntimeError

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction

_ACH_DEPOSIT: str = "Ach Deposit"
_ACH_WITHDRAWAL: str = "Ach Withdrawal"
_BUY_CURRENCY: str = "Buy Currency"
_BUY_QUANTITY: str = "Buy Quantity"
_CRYPTO_TRANSFER: str = "Crypto Transfer"
_DATE: str = "Date"
_INTEREST_PAYMENT: str = "Interest Payment"
_REFERRAL_BONUS: str = "Referral Bonus"
_BONUS_PAYMENT = "Bonus Payment"
_SOLD_CURRENCY: str = "Sold Currency"
_SOLD_QUANTITY: str = "Sold Quantity"
_TRADE: str = "Trade"
_TRADE_ID: str = "Trade ID"
_TYPE: str = "Type"
_WITHDRAWAL: str = "Withdrawal"
_BIA_WITHDRAWAL: str = "BIA Withdraw"
_WITHDRAWAL_FEE: str = "Withdrawal Fee"


class InputPlugin(AbstractInputPlugin):

    __BLOCKFI: str = "BlockFi"

    __CURRENCY_INDEX: int = 0
    __AMOUNT_INDEX: int = 1
    __TYPE_INDEX: int = 2
    __TIMESTAMP_INDEX: int = 3
    #    __SPOT_PRICE_INDEX: int = 4 # It used to be present in the CSV, but sometime in 2021 BlockFi removed this field.

    __DELIMITER = ","

    def __init__(
        self,
        account_holder: str,
        transaction_csv_file: str,
        trade_csv_file: Optional[str] = None,
        native_fiat: Optional[str] = None,
    ) -> None:

        super().__init__(account_holder=account_holder, native_fiat=native_fiat)
        self.__transaction_csv_file: str = transaction_csv_file
        self.__trade_csv_file: Optional[str] = trade_csv_file
        self.__logger: logging.Logger = create_logger(f"{self.__BLOCKFI}/{self.account_holder}")

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        last_withdrawal_fee: Optional[RP2Decimal] = None
        with open(self.__transaction_csv_file, encoding="utf-8") as transaction_csv_file:
            lines = reader(transaction_csv_file)
            # Skip header line
            header = next(lines)
            self.__logger.debug("Header: %s", header)
            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                self.__logger.debug("Transaction: %s", raw_data)

                if last_withdrawal_fee is not None and line[self.__TYPE_INDEX] != _WITHDRAWAL:
                    raise RP2RuntimeError(f"Internal error: withdrawal fee {last_withdrawal_fee} is not followed by withdrawal")

                transaction_type: str = line[self.__TYPE_INDEX]
                if transaction_type == _INTEREST_PAYMENT:
                    last_withdrawal_fee = None
                    result.append(
                        InTransaction(
                            plugin=self.__BLOCKFI,
                            unique_id=Keyword.UNKNOWN.value,
                            raw_data=raw_data,
                            timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            asset=line[self.__CURRENCY_INDEX],
                            exchange=self.__BLOCKFI,
                            holder=self.account_holder,
                            transaction_type=Keyword.INTEREST.value,
                            spot_price=Keyword.UNKNOWN.value,
                            crypto_in=line[self.__AMOUNT_INDEX],
                            fiat_fee="0",
                        )
                    )
                elif transaction_type in [_REFERRAL_BONUS, _BONUS_PAYMENT]:
                    last_withdrawal_fee = None
                    result.append(
                        InTransaction(
                            plugin=self.__BLOCKFI,
                            unique_id=Keyword.UNKNOWN.value,
                            raw_data=raw_data,
                            timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            asset=line[self.__CURRENCY_INDEX],
                            exchange=self.__BLOCKFI,
                            holder=self.account_holder,
                            transaction_type=Keyword.INCOME.value,
                            spot_price=Keyword.UNKNOWN.value,
                            crypto_in=line[self.__AMOUNT_INDEX],
                            fiat_fee="0",
                            notes="Referral Bonus",
                        )
                    )
                elif transaction_type == _CRYPTO_TRANSFER:
                    last_withdrawal_fee = None
                    result.append(
                        IntraTransaction(
                            plugin=self.__BLOCKFI,
                            unique_id=Keyword.UNKNOWN.value,
                            raw_data=raw_data,
                            timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            asset=line[self.__CURRENCY_INDEX],
                            from_exchange=Keyword.UNKNOWN.value,
                            from_holder=Keyword.UNKNOWN.value,
                            to_exchange=self.__BLOCKFI,
                            to_holder=self.account_holder,
                            spot_price=Keyword.UNKNOWN.value,
                            crypto_sent=Keyword.UNKNOWN.value,
                            crypto_received=line[self.__AMOUNT_INDEX],
                        )
                    )
                elif transaction_type == _ACH_WITHDRAWAL:
                    last_withdrawal_fee = None
                    result.append(
                        OutTransaction(
                            plugin=self.__BLOCKFI,
                            unique_id=Keyword.UNKNOWN.value,
                            raw_data=raw_data,
                            timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            asset=line[self.__CURRENCY_INDEX],
                            exchange=self.__BLOCKFI,
                            holder=self.account_holder,
                            transaction_type=Keyword.SELL.value,
                            spot_price=Keyword.UNKNOWN.value,
                            crypto_out_no_fee=str(-RP2Decimal(line[self.__AMOUNT_INDEX])),
                            crypto_fee="0",
                            notes="ACH withdrawal",
                        )
                    )
                elif transaction_type == _WITHDRAWAL:
                    amount: RP2Decimal = RP2Decimal(line[self.__AMOUNT_INDEX])
                    amount = -amount  # type: ignore
                    if last_withdrawal_fee is not None:
                        amount += last_withdrawal_fee
                    last_withdrawal_fee = None
                    result.append(
                        IntraTransaction(
                            plugin=self.__BLOCKFI,
                            unique_id=Keyword.UNKNOWN.value,
                            raw_data=raw_data,
                            timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            asset=line[self.__CURRENCY_INDEX],
                            from_exchange=self.__BLOCKFI,
                            from_holder=self.account_holder,
                            to_exchange=Keyword.UNKNOWN.value,
                            to_holder=Keyword.UNKNOWN.value,
                            spot_price=Keyword.UNKNOWN.value,
                            crypto_sent=str(amount),
                            crypto_received=Keyword.UNKNOWN.value,
                        )
                    )
                elif transaction_type == _ACH_DEPOSIT:
                    last_withdrawal_fee = None
                    result.append(
                        InTransaction(
                            plugin=self.__BLOCKFI,
                            unique_id=Keyword.UNKNOWN.value,
                            raw_data=raw_data,
                            timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            asset=line[self.__CURRENCY_INDEX],
                            exchange=self.__BLOCKFI,
                            holder=self.account_holder,
                            transaction_type=Keyword.BUY.value,
                            spot_price=Keyword.UNKNOWN.value,
                            crypto_in=line[self.__AMOUNT_INDEX],
                            fiat_fee="0",
                            notes="ACH deposit",
                        )
                    )
                elif transaction_type == _TRADE:
                    # Trades will be handled by parsing trade_report_all.csv
                    # export
                    continue
                elif transaction_type == _BIA_WITHDRAWAL:
                    # these withdrawals are internal transfers within blockfi which is why they are skipped
                    # https://github.com/eprbell/dali-rp2/pull/64
                    self.__logger.debug("BIA Withdraw: %s", raw_data)
                    continue
                elif transaction_type == _WITHDRAWAL_FEE:
                    last_withdrawal_fee = RP2Decimal(line[self.__AMOUNT_INDEX])
                    last_withdrawal_fee = -last_withdrawal_fee  # type: ignore
                else:
                    self.__logger.error("Unsupported transaction type (skipping): %s. Please open an issue at %s", raw_data, self.ISSUES_URL)

        if self.__trade_csv_file:
            result += self.parse_trade_report(self.__trade_csv_file)
        return result

    def parse_trade_report(self, file_path: str) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        with open(file_path, encoding="utf-8") as csv_file:
            lines = reader(csv_file)
            # skip csv header
            header = next(lines)
            self.__logger.debug("Header: %s", header)

            column_index: Dict[str, int] = {}
            for (index, name) in enumerate(header):
                column_index[name] = index

            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                self.__logger.debug("Transaction: %s", raw_data)

                transaction_type: str = line[column_index[_TYPE]]
                if transaction_type != "Trade":
                    raise RP2RuntimeError(f"Internal error: unsupported transaction type: {transaction_type}")

                trade_id: str = line[column_index[_TRADE_ID]]
                date: str = line[column_index[_DATE]]
                timestamp: str = f"{date} -00:00"
                from_currency: str = line[column_index[_SOLD_CURRENCY]].upper()
                from_size: str = line[column_index[_SOLD_QUANTITY]]
                to_currency: str = line[column_index[_BUY_CURRENCY]].upper()
                to_size: str = line[column_index[_BUY_QUANTITY]]

                result.append(
                    OutTransaction(
                        plugin=self.__BLOCKFI,
                        unique_id=trade_id,
                        raw_data=raw_data,
                        timestamp=timestamp,
                        asset=from_currency,
                        exchange=self.__BLOCKFI,
                        holder=self.account_holder,
                        transaction_type="Sell",
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_out_no_fee=from_size,
                        crypto_fee="0",
                        notes=f"Sell side of trade: {from_size} {from_currency} -> {to_size} {to_currency}",
                    ),
                )
                result.append(
                    InTransaction(
                        plugin=self.__BLOCKFI,
                        unique_id=f"{trade_id}/buy",
                        raw_data=raw_data,
                        timestamp=timestamp,
                        asset=to_currency,
                        exchange=self.__BLOCKFI,
                        holder=self.account_holder,
                        transaction_type="Buy",
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_in=to_size,
                        fiat_fee="0",
                        notes=f"Buy side of trade: {from_size} {from_currency} -> {to_size} {to_currency}",
                    ),
                )

        return result



# Copyright 2023 jamesbaber1
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Optional, cast

from rp2.abstract_country import AbstractCountry
from rp2.configuration import Configuration, MAX_DATE, MIN_DATE
from rp2.in_transaction import InTransaction as RP2InTransaction
from rp2.input_data import InputData
from rp2.intra_transaction import IntraTransaction as RP2IntraTransaction
from rp2.logger import create_logger
from rp2.ods_parser import parse_ods, open_ods
from rp2.out_transaction import OutTransaction as RP2OutTransaction

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction


class InputPlugin(AbstractInputPlugin):
    __RP2_INPUT: str = "RP2 Input"

    def __init__(
        self,
        configuration_path: str,
        input_file: str,
        native_fiat: Optional[str] = None,
    ) -> None:
        super().__init__(account_holder="", native_fiat=native_fiat)
        self.__configuration_path: str = configuration_path
        self.__input_file: str = input_file
        self.__logger: logging.Logger = create_logger(self.__RP2_INPUT)

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        rp2_configuration: Configuration = Configuration(
            configuration_path=self.__configuration_path,
            country=country,
            from_date=MIN_DATE,
            to_date=MAX_DATE,
        )

        input_file_handle: object = open_ods(configuration=rp2_configuration, input_file_path=self.__input_file)
        assets = sorted(list(rp2_configuration.assets))

        for asset in assets:
            self.__logger.info("Processing %s", asset)

            input_data: InputData = parse_ods(
                configuration=rp2_configuration,
                asset=asset,
                input_file_handle=input_file_handle
            )
            self.__logger.debug("InputData object: %s", input_data)
            for asset_entry in input_data.unfiltered_in_transaction_set:
                in_transaction: RP2InTransaction = cast(RP2InTransaction, asset_entry)
                self.__logger.debug("Transaction: %s", str(in_transaction))
                result.append(InTransaction(
                    plugin=self.__RP2_INPUT,
                    unique_id=in_transaction.unique_id,
                    raw_data=str(in_transaction),
                    timestamp=str(in_transaction.timestamp),
                    asset=in_transaction.asset,
                    exchange=in_transaction.exchange,
                    holder=in_transaction.holder,
                    transaction_type=in_transaction.transaction_type.value,
                    spot_price=str(in_transaction.spot_price),
                    crypto_in=str(in_transaction.crypto_in),
                    crypto_fee=str(in_transaction.crypto_fee) if in_transaction.crypto_fee else None,
                    fiat_in_no_fee=str(in_transaction.fiat_in_no_fee) if in_transaction.fiat_in_no_fee else None,
                    fiat_in_with_fee=str(in_transaction.fiat_in_with_fee) if in_transaction.fiat_in_with_fee else None,
                    fiat_fee=str(in_transaction.fiat_fee) if in_transaction.fiat_fee else None,
                    notes=str(in_transaction.notes) if in_transaction.notes else None
                ))

            for asset_transfer in input_data.unfiltered_intra_transaction_set:
                intra_transaction: RP2IntraTransaction = cast(RP2IntraTransaction, asset_transfer)
                self.__logger.debug("Transaction: %s", str(intra_transaction))
                result.append(
                    IntraTransaction(
                        plugin=self.__RP2_INPUT,
                        unique_id=intra_transaction.unique_id,
                        raw_data=str(intra_transaction),
                        timestamp=str(intra_transaction.timestamp),
                        asset=intra_transaction.asset,
                        from_exchange=intra_transaction.from_exchange,
                        from_holder=intra_transaction.from_holder,
                        to_exchange=intra_transaction.to_exchange,
                        to_holder=intra_transaction.to_holder,
                        spot_price=str(intra_transaction.spot_price) if intra_transaction.spot_price else None,
                        crypto_sent=str(intra_transaction.crypto_sent),
                        crypto_received=str(intra_transaction.crypto_received),
                        notes=str(intra_transaction.notes) if intra_transaction.notes else None
                    ))

            for asset_exit in input_data.unfiltered_out_transaction_set:
                out_transaction: RP2OutTransaction = cast(RP2OutTransaction, asset_exit)
                self.__logger.debug("Transaction: %s", str(out_transaction))
                result.append(
                    OutTransaction(
                        plugin=self.__RP2_INPUT,
                        unique_id=out_transaction.unique_id,
                        raw_data=str(out_transaction),
                        timestamp=str(out_transaction.timestamp),
                        asset=out_transaction.asset,
                        exchange=out_transaction.exchange,
                        holder=out_transaction.holder,
                        transaction_type=out_transaction.transaction_type.value,
                        spot_price=str(out_transaction.spot_price),
                        crypto_out_no_fee=str(out_transaction.crypto_out_no_fee),
                        crypto_fee=str(out_transaction.crypto_fee),
                        crypto_out_with_fee=str(out_transaction.crypto_out_with_fee),
                        fiat_out_no_fee=str(out_transaction.fiat_out_no_fee) if out_transaction.fiat_out_no_fee else None,
                        fiat_fee=str(out_transaction.fiat_fee) if out_transaction.fiat_fee else None,
                        notes=str(out_transaction.notes) if out_transaction.notes else None
                    ))
        return result



# Copyright 2022 Steve Davis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import FrozenInstanceError
from datetime import datetime, timedelta, timezone

import pytest
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2ValueError

from dali.configuration import Keyword
from dali.historical_bar import HistoricalBar

# Test values
BAR_SECONDS: int = 60
BAR_START_TIMESTAMP: datetime = datetime(2022, 1, 1, hour=13, minute=30).replace(tzinfo=timezone.utc)
BAR_OPEN: RP2Decimal = RP2Decimal("1.0")
BAR_HIGH: RP2Decimal = RP2Decimal("3.0")
BAR_LOW: RP2Decimal = RP2Decimal("0.5")
BAR_CLOSE: RP2Decimal = RP2Decimal("2.0")
DONT_CARE_TIMESTAMP = datetime(2022, 2, 2, hour=2, minute=2).replace(tzinfo=timezone.utc)


@pytest.fixture  # type: ignore
def bar() -> HistoricalBar:  # pylint: disable=C0104, W0621
    """Create a HistoricalBar to test with."""
    # Construct a hardcoded dataframe matching the type returned by Historic_Crypto.HistoricalData.
    return HistoricalBar(
        duration=timedelta(seconds=BAR_SECONDS),
        timestamp=BAR_START_TIMESTAMP,
        open=BAR_OPEN,
        high=BAR_HIGH,
        low=BAR_LOW,
        close=BAR_CLOSE,
        volume=ZERO,
    )


def test_historical_bar_is_frozen(bar: HistoricalBar) -> None:  # pylint: disable=C0104, W0621
    """Verify HistoricalBar data elements cannot be changed."""
    # These tests will pass for either NamedTuple rasing AttributeError or frozen dataclass raising FrozenInstanceError
    with pytest.raises((AttributeError, FrozenInstanceError)):
        bar.duration = None  # type: ignore
    with pytest.raises((AttributeError, FrozenInstanceError)):
        bar.timestamp = None  # type: ignore
    with pytest.raises((AttributeError, FrozenInstanceError)):
        bar.open = None  # type: ignore
    with pytest.raises((AttributeError, FrozenInstanceError)):
        bar.high = None  # type: ignore
    with pytest.raises((AttributeError, FrozenInstanceError)):
        bar.low = None  # type: ignore
    with pytest.raises((AttributeError, FrozenInstanceError)):
        bar.close = None  # type: ignore
    with pytest.raises((AttributeError, FrozenInstanceError)):
        bar.volume = None  # type: ignore


def test_derive_transaction_price_invalid(bar: HistoricalBar) -> None:  # pylint: disable=C0104, W0621
    """Verify invalid price selection parameter raises an exception."""
    with pytest.raises(RP2ValueError):
        bar.derive_transaction_price(DONT_CARE_TIMESTAMP, "Invalid")


def test_derive_transaction_price_open(bar: HistoricalBar) -> None:  # pylint: disable=C0104, W0621
    assert BAR_OPEN == bar.derive_transaction_price(DONT_CARE_TIMESTAMP, Keyword.HISTORICAL_PRICE_OPEN.value)


def test_derive_transaction_price_high(bar: HistoricalBar) -> None:  # pylint: disable=C0104, W0621
    assert BAR_HIGH == bar.derive_transaction_price(DONT_CARE_TIMESTAMP, Keyword.HISTORICAL_PRICE_HIGH.value)


def test_derive_transaction_price_low(bar: HistoricalBar) -> None:  # pylint: disable=C0104, W0621
    assert BAR_LOW == bar.derive_transaction_price(DONT_CARE_TIMESTAMP, Keyword.HISTORICAL_PRICE_LOW.value)


def test_derive_transaction_price_close(bar: HistoricalBar) -> None:  # pylint: disable=C0104, W0621
    assert BAR_CLOSE == bar.derive_transaction_price(DONT_CARE_TIMESTAMP, Keyword.HISTORICAL_PRICE_CLOSE.value)


@pytest.mark.parametrize("seconds", [-1, 0, 1, int(BAR_SECONDS / 2) - 1])  # test various time offsets after bar start timestamp
def test_derive_transaction_price_nearest_open(bar: HistoricalBar, seconds: int) -> None:  # pylint: disable=C0104, W0621
    transaction_timestamp = (BAR_START_TIMESTAMP + timedelta(seconds=seconds)).replace(tzinfo=timezone.utc)
    assert BAR_OPEN == bar.derive_transaction_price(transaction_timestamp, Keyword.HISTORICAL_PRICE_NEAREST.value)


@pytest.mark.parametrize(  # test various time offsets after bar start timestamp
    "seconds", [int(BAR_SECONDS / 2) + 1, BAR_SECONDS - 1, BAR_SECONDS, BAR_SECONDS + 1]
)
def test_derive_transaction_price_nearest_close(bar: HistoricalBar, seconds: int) -> None:  # pylint: disable=C0104, W0621
    transaction_timestamp = (BAR_START_TIMESTAMP + timedelta(seconds=seconds)).replace(tzinfo=timezone.utc)
    assert BAR_CLOSE == bar.derive_transaction_price(transaction_timestamp, Keyword.HISTORICAL_PRICE_NEAREST.value)



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Withdrawal CSV Format: 日時 (timestamp), 数量 (amount), 手数料 (transaction fee), 合計 (total), ラベル (user-provided label), アドレス (address), Txid, ステータス (status)

import logging
from csv import reader
from datetime import datetime
from datetime import timezone as DatetimeTimezone
from typing import List, Optional

from pytz import timezone as PytzTimezone

from rp2.abstract_country import AbstractCountry
from rp2.logger import create_logger
from rp2.rp2_error import RP2ValueError

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction


class InputPlugin(AbstractInputPlugin):

    __BITBANK: str = "Bitbank.cc"
    __BITBANK_PLUGIN: str = "Bitbank_Supplemental_CSV"

    __TIMESTAMP_INDEX: int = 0
    __SENT_AMOUNT: int = 1
    __DEPOSIT_TOTAL: int = 1
    __TRANSACTION_FEE: int = 2
    __DEPOSIT_STATUS: int = 2
    __TOTAL: int = 3
    __LABEL: int = 4
    __ADDRESS: int = 5
    __TX_ID: int = 6
    __STATUS: int = 7

    __DELIMITER: str = ","

    def __init__(
        self,
        account_holder: str,
        deposits_csv_file: Optional[str] = None,
        deposits_code: Optional[str] = None,
        withdrawals_csv_file: Optional[str] = None,
        withdrawals_code: Optional[str] = None,
        native_fiat: Optional[str] = None,
    ) -> None:

        super().__init__(account_holder=account_holder, native_fiat=native_fiat)
        self.__deposits_csv_file: Optional[str] = deposits_csv_file
        self.__deposits_code: Optional[str] = deposits_code
        self.__withdrawals_csv_file: Optional[str] = withdrawals_csv_file
        # Code of the asset being withdrawn since it is NOT included in the CSV file.
        self.__withdrawals_code: Optional[str] = withdrawals_code
        self.__logger: logging.Logger = create_logger(f"{self.__BITBANK_PLUGIN}/{self.account_holder}")

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        if self.__withdrawals_csv_file:
            result.extend(self.parse_withdrawals_file(self.__withdrawals_csv_file))

        if self.__deposits_csv_file:
            result.extend(self.parse_deposits_file(self.__deposits_csv_file))

        return result

    def parse_deposits_file(self, file_path: str) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []
        if not self.__deposits_code:
            raise RP2ValueError("Bitbank.cc supplemental plugin deposits file declared without deposits code.")

        with open(file_path, encoding="utf-8") as csv_file:
            lines = reader(csv_file)

            header = next(lines)
            self.__logger.debug("Header: %s", header)
            for line in lines:
                if line[self.__DEPOSIT_STATUS] == "DONE":
                    raw_data: str = self.__DELIMITER.join(line)
                    self.__logger.debug("Transaction: %s", raw_data)

                    jst_timezone = PytzTimezone("Asia/Tokyo")
                    jst_datetime: datetime = jst_timezone.localize(datetime.strptime(line[self.__TIMESTAMP_INDEX], "%Y/%m/%d %H:%M:%S"))
                    utc_timestamp: str = jst_datetime.astimezone(DatetimeTimezone.utc).strftime("%Y-%m-%d %H:%M:%S%z")

                    result.append(
                        InTransaction(
                            plugin=self.__BITBANK_PLUGIN,
                            unique_id=Keyword.UNKNOWN.value,
                            raw_data=raw_data,
                            timestamp=utc_timestamp,
                            asset=self.__deposits_code,
                            exchange=self.__BITBANK,
                            holder=self.account_holder,
                            transaction_type=Keyword.BUY.value,
                            spot_price="1.0",
                            crypto_in=str(line[self.__DEPOSIT_TOTAL]),
                            crypto_fee=None,
                            fiat_in_no_fee=str(line[self.__DEPOSIT_TOTAL]),
                            fiat_in_with_fee=str(line[self.__DEPOSIT_TOTAL]),
                            fiat_ticker="JPY",
                            notes="Fiat deposit",
                        )
                    )

        return result

    def parse_withdrawals_file(self, file_path: str) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []
        if not self.__withdrawals_code:
            raise RP2ValueError("Bitbank.cc supplemental plugin withdrawals file declared without withdrawals code.")

        with open(file_path, encoding="utf-8") as csv_file:
            lines = reader(csv_file)

            header = next(lines)
            self.__logger.debug("Header: %s", header)
            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                self.__logger.debug("Transaction: %s", raw_data)

                jst_timezone = PytzTimezone("Asia/Tokyo")
                jst_datetime: datetime = jst_timezone.localize(datetime.strptime(line[self.__TIMESTAMP_INDEX], "%Y/%m/%d %H:%M:%S"))
                utc_timestamp: str = jst_datetime.astimezone(DatetimeTimezone.utc).strftime("%Y-%m-%d %H:%M:%S%z")

                result.append(
                    IntraTransaction(
                        plugin=self.__BITBANK_PLUGIN,
                        unique_id=line[self.__TX_ID],
                        raw_data=raw_data,
                        timestamp=utc_timestamp,
                        asset=self.__withdrawals_code,
                        from_exchange=self.__BITBANK,
                        from_holder=self.account_holder,
                        to_exchange=Keyword.UNKNOWN.value,
                        to_holder=Keyword.UNKNOWN.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_sent=str(line[self.__TOTAL]),
                        crypto_received=Keyword.UNKNOWN.value,
                    )
                )

        return result
